#!/usr/bin/env python
import sys, os, stat, fijibin
from io import BytesIO
from zipfile import ZipFile
try:
    from zipfile import BadZipZile
except ImportError:
    from zipfile import BadZipfile
    BadZipFile = BadZipfile


def fetch():
    "Fetch and extract latest Life-Line version of Fiji is just ImageJ to `~/.bin`."
    if os.path.isdir(fijibin.FIJI_ROOT):
        print('%s already exists, not downloading.' % fijibin.FIJI_ROOT)
        return

    try:
        # python 2
        from urllib2 import urlopen, HTTPError, URLError
    except ImportError:
        # python 3
        from urllib.request import urlopen, HTTPError, URLError

    url = fijibin.URL

    try:
        print('Getting fiji from %s' % url)
        req = urlopen(url)
        try:
            size = int(req.info()['content-length'])
        except AttributeError:
            size = -1

        chunk = 512*1024
        fp = BytesIO()
        i = 0
        while 1:
            data = req.read(chunk)
            if not data:
                break
            fp.write(data)
            if size > 0:
                percent = fp.tell() // (size/100)
                msg = 'Downloaded %d percent      \r' % percent
            else:
                msg = 'Downloaded %d bytes\r' % fp.tell()
            sys.stdout.write(msg)
    except (HTTPError, URLError) as e:
        print('Error getting fiji: {}'.format(e))
        sys.exit(1)

    try:
        print('\nExtracting zip')
        z = ZipFile(fp)
        z.extractall(fijibin.BIN_FOLDER)
        # move to Fiji-VERSION.app to easily check if it exists (upon fijibin upgrade)
        os.rename(fijibin.EXTRACT_FOLDER, fijibin.FIJI_ROOT)
    except (BadZipFile, IOError) as e:
        print('Error extracting zip: {}'.format(e))
        sys.exit(1)

    for path in fijibin.BIN_NAMES.values():
        st = os.stat(path)
        os.chmod(path, st.st_mode | stat.S_IEXEC)



"""
Latest Life-Line version of fiji for easy inclusion in Python projects.
"""
__all__ = []
from os.path import join, dirname
VERSION = open(join(dirname(__file__), 'VERSION')).read().strip()
FIJI_VERSION = '20141125'
BASE_URL = 'http://fiji.sc/downloads/Life-Line/fiji-'
END_URL = '-' + FIJI_VERSION + '.zip'

import platform, os

HOME = os.path.expanduser('~')
BIN_FOLDER = os.path.join(HOME, '.bin')
EXTRACT_FOLDER = os.path.join(BIN_FOLDER, 'Fiji.app')
FIJI_ROOT = os.path.join(BIN_FOLDER, 'Fiji-' + FIJI_VERSION + '.app')
BIN_NAMES = {
    'Darwin64bit': os.path.join(FIJI_ROOT, 'Contents/MacOS/ImageJ-macosx'),
    'Linux32bit': os.path.join(FIJI_ROOT, 'ImageJ-linux32'),
    'Linux64bit': os.path.join(FIJI_ROOT, 'ImageJ-linux64'),
    'Windows32bit': os.path.join(FIJI_ROOT, 'ImageJ-win32.exe'),
    'Windows64bit': os.path.join(FIJI_ROOT, 'ImageJ-win64.exe')
}

_unix_url = BASE_URL + 'nojre' + END_URL
URLS = {
    'Darwin64bit': _unix_url,
    'Linux32bit': _unix_url,
    'Linux64bit': _unix_url,
    'Windows32bit': BASE_URL + 'win32' + END_URL,
    'Windows64bit': BASE_URL + 'win64'  + END_URL
}

SYSTEM = platform.system() + platform.architecture()[0]
BIN = BIN_NAMES[SYSTEM]
URL = URLS[SYSTEM]



# encoding: utf-8
"""
Run Fiji is just ImageJ macros headless with python.
"""
import pydebug, subprocess, os, fijibin, re
from tempfile import mkstemp

# debug with DEBUG=fijibin python script.py
debug = pydebug.debug('fijibin')

_bin = fijibin.BIN

##
# Running macros
##
def run(macro, output_files=[], force_close=True):
    """
    Runs Fiji with the suplied macro. Output of Fiji can be viewed by
    setting environment variable `DEBUG=fijibin`.

    Parameters
    ----------
    macro : string or list of strings
        IJM-macro(s) to run. If list of strings, it will be joined with
        a space, so all statements should end with ``;``.
    output_files : list
        Files to check if exists after macro has been run. Files specified that
        do not exist after macro is done will print a warning message.
    force_close : bool
        Will add ``eval("script", "System.exit(42);");`` to end of macro. Exit
        code 42 is used to overcome that errors in macro efficiently will exit
        Fiji with error code 0. In other words, if this line in the macro is
        reached, the macro has most probably finished without errors. This
        is the default behaviour.

        One should also note that Fiji doesn't terminate right away if
        ``System.exit()`` is left out, and it may take several minutes for
        Fiji to close.

    Returns
    -------
    int
        Files from output_files which exists after running macro.
    """
    if type(macro) == list:
        macro = ' '.join(macro)
    if len(macro) == 0:
        print('fijibin.macro.run got empty macro, not starting fiji')
        return _exists(output_files)
    if force_close:
        # make sure fiji halts immediately when done
        # hack: use error code 42 to check if macro has run sucessfully
        macro = macro + 'eval("script", "System.exit(42);");'

    # escape backslashes (windows file names)
    #                 not \ \  not \      g1 \\ g2
    macro = re.sub(r"([^\\])\\([^\\])", r"\1\\\\\2", macro)

    debug('macro {}'.format(macro))

    # avoid verbose output of Fiji when DEBUG environment variable set
    env = os.environ.copy()
    debugging = False
    if 'DEBUG' in env:
        if env['DEBUG'] == 'fijibin' or env['DEBUG'] == '*':
            debugging = True
        del env['DEBUG']

    fptr, temp_filename = mkstemp(suffix='.ijm')
    m = os.fdopen(fptr, mode='w')
    m.write(macro)
    m.flush() # make sure macro is written before running Fiji
    m.close()

    cmd = [_bin, '--headless', '-macro', temp_filename]
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE, env=env)
    out, err = proc.communicate()

    for line in out.decode('latin1', errors='ignore').splitlines():
        debug('stdout:' + line)
    for line in err.decode('latin1', errors='ignore').splitlines():
        debug('stderr:' + line)

    if force_close and proc.returncode != 42:
        print('fijibin ERROR: Fiji did not successfully ' +
              'run macro {}'.format(temp_filename))
        if not debugging:
            print('fijibin Try running script with ' +
                  '`DEBUG=fijibin python your_script.py`')
    else:
        # only delete if everything is ok
        os.remove(temp_filename)


    # return output_files which exists
    return _exists(output_files)

##
# Collection of macros
##
def stitch(folder, filenames, x_size, y_size, output_filename,
                 x_start=0, y_start=0, overlap=10):
    """
    Creates a Fiji Grid/Collection stitching macro. Parameters are the same as
    in the plugin and are described in further detail here:
    http://fiji.sc/Image_Stitching#Grid.2FCollection_Stitching.

    **Default stitch parameters:**

    * Filename defined positions
    * Compute overlap
    * Subpixel accurancy
    * Save computation time (but use more RAM)
    * Fusion method: Linear blending
    * Regression threshold: 0.30
    * Max/avg displacement threshold: 2.50
    * Absolute displacement threshold: 3.50


    Parameters
    ----------
    folder : string
        Path to folder with images or folders with images.
        Example: */path/to/slide--S00/chamber--U01--V02/*
    filenames : string
        Filenames of images.
        Example: *field-X{xx}-Y{yy}/image-X{xx}-Y{yy}.ome.tif*
    x_size : int
        Size of grid, number of images in x direction.
    y_size : int
        Size of grid, number of images in y direction.
    output_filename : string
        Where to store fused image. Should be `.png`.
    x_start : int
        Which x position grid start with.
    y_start : int
        Which y position grid start with.
    overlap : number
        Tile overlap in percent. Fiji will find the optimal overlap, but a
        precise overlap assumption will decrase computation time.

    Returns
    -------
    string
        IJM-macro.
    """

    macro = []
    macro.append('run("Grid/Collection stitching",')
    macro.append('"type=[Filename defined position]')
    macro.append('order=[Defined by filename         ]')
    macro.append('grid_size_x={}'.format(x_size))
    macro.append('grid_size_y={}'.format(y_size))
    macro.append('tile_overlap={}'.format(overlap))
    macro.append('first_file_index_x={}'.format(x_start))
    macro.append('first_file_index_y={}'.format(y_start))
    macro.append('directory=[{}]'.format(folder))
    macro.append('file_names=[{}]'.format(filenames))
    macro.append('output_textfile_name=TileConfiguration.txt')
    macro.append('fusion_method=[Linear Blending]')
    macro.append('regression_threshold=0.30')
    macro.append('max/avg_displacement_threshold=2.50')
    macro.append('absolute_displacement_threshold=3.50')
    macro.append('compute_overlap')
    macro.append('subpixel_accuracy')
    macro.append('computation_parameters=[Save computation time (but use more RAM)]')
    # use display, such that we can specify output filename
    # this is 'Fused and display' for previous stitching version!!
    macro.append('image_output=[Fuse and display]");')
    # save to png
    macro.append('selectWindow("Fused");')
    macro.append('saveAs("PNG", "{}");'.format(output_filename))
    macro.append('close();')

    return ' '.join(macro)


##
# Helper functions
##
def _exists(filenames):
    """Check if every filename exists. If not, print an error
    message and remove the item from the list.

    Parameters
    ----------
    filenames : list
        List of filenames to check for existence.

    Returns
    -------
    list
        Filtered list of filenames that exists.
    """
    exists = []
    for filename in filenames:
        if os.path.isfile(filename):
            exists.append(filename)
        else:
            print('fijibin ERROR missing output file {}'.format(filename))

    return exists



#!/usr/bin/env python
# encoding: utf-8

# readme
import os
if os.path.exists('README.rst'):
    long_description = open('README.rst').read()
else:
    print('README.rst missing')
    long_description = ''

from setuptools import setup, find_packages
from setuptools.command.install import install
from fijibin.fetch import fetch

class CustomInstall(install):
    """Fetch Fiji and install package."""
    def run(self):
        fetch()
        install.run(self)

setup(name='fijibin',
      version=open(os.path.join('fijibin', 'VERSION')).read().strip(),
      description='Latest Life-Line version of Fiji for easy inclusion in python projects.',
      author='Arve Seljebu',
      author_email='arve.seljebu@gmail.com',
      url='https://github.com/arve0/fijibin',
      license='MIT',
      packages=['fijibin'],
      install_requires=[
          'pydebug'
      ],
      package_data={'fijibin': ['VERSION']},
      include_package_data=True,
      long_description=long_description,
      cmdclass={'install': CustomInstall} )
